/* \TEX{
   At the compilation time, a test program is
   compiled. To use it, you only have to type |bin/arch/zentest|. This
   program tests all the functions of the library (and so can take
   time ...).

   The options of this program are described by |bin/arch/zentest -h|.

   The five sets of procedures are tested. That is to say:
   \begin{itemize}
   \item Functions on elements.
   \item Functions on polynomials.
   \item Functions on matrices.
   \item Functions on series.
   \item Functions on elliptic curves.
   \end{itemize}

   The procedures of this program are not in |libzen.a| but
   can be good examples of procedures written with |ZEN| functions.

   The same tests are performed for different finite ring following user
   options specified at the execution of the process (see |Usage()|).
}
   */
#include <stdio.h>
#include "zen.h"
#include "zef.h"
#include "test.h"
#include "zenfact.h"

#define FINITE_FIELD TRUE

#define NB_TEST_POLY_FACTOR 10
#define NB_TEST_POLY_ROOTS 10
#define NB_TEST_POLY_IRREDUCIBLE 5
#define DEG_POLY 5
#define ZEN_DEG_EXT 2

extern void Usage ___(());

#include "zentest_data.h"


void main(argc, argv)
     int argc;
     char *argv[];
{
  char **def;
  int c, i, rg, nb_rg;
  int lib_flag, proc_flag;
  int zen2lib, zen2lib0, zen2lib1, zen2lib2;
  int zepslib, zepslib0, zepslib1, zepslib2;
  int zetablib, zetablib0, zetablib1, zetablib2;
  int zeplib, zeplib0, zeplib1, zeplib2;
  int zeflib, zeflib0, zeflib1, zeflib2;
  int level[NB_RG];
  double tt, TT;
  BigNum n;
  BigNumLength nl;
  int first;
  ZENRing Rg[NB_RG], oldrg, RgClone;
  ZENPoly P1;
  ZENPrc prc;
  ZENCln cln;
  
  extern char *optarg;

  zbnsrandom(1);

  setbuf(stdout, NULL);
  setbuf(stderr, NULL);

  def = NULL;
  lib_flag = proc_flag = 1;
  zen2lib = zen2lib0 = zen2lib1 = zen2lib2 = 1; 
  zepslib = zepslib0 = zepslib1 = zepslib2 = 1; 
  zetablib = zetablib0 = zetablib1 = zetablib2 = 1; 
  zeplib = zeplib0 = zeplib1 = zeplib2 = 1;
  zeflib = zeflib0 = APPROXIMATE_RESULTS;
  zeflib1 = zeflib2 = 0;
  
  while ((c = getopt(argc, argv, "l:p:")) != EOF)
    switch (c) {
      
      /* The finite rings to test */
    case 'l':
      
      if (lib_flag) {
        lib_flag = 0;
        zen2lib = zen2lib0 = zen2lib1 = zen2lib2 = 0;
        zepslib = zepslib0 = zepslib1 = zepslib2 = 0;
        zetablib = zetablib0 = zetablib1 = zetablib2 = 0;
        zeplib = zeplib0 = zeplib1 = zeplib2 = 0;
        zeflib = zeflib0 = zeflib1 = zeflib2 = 0;
      }
      switch(optarg[0]) {
      case 't':
        if (!strcmp(optarg, "t")) {
          zen2lib0 = zen2lib1 = zen2lib2 = 1;
          zen2lib = 1;
          break;
        }
        else if (!strcmp(optarg, "t0")) {
          zen2lib0 = 1;
          zen2lib = 1;
          break;
        }
        else if (!strcmp(optarg, "t1")) {
          zen2lib1 = 1;
          zen2lib = 1;
          break;
        }
        else if (!strcmp(optarg, "t2")) {
          zen2lib2 = 1;
          zen2lib = 1;
          break;
        }
        break;
        
      case 's':
        if (!strcmp(optarg, "s")) {
          zepslib0 = zepslib1 = zepslib2 = 1;
          zepslib = 1;
          break;
        }
        else if (!strcmp(optarg, "s0")) {
          zepslib0 = 1;
          zepslib = 1;
          break;
        }
        else if (!strcmp(optarg, "s1")) {
          zepslib1 = 1;
          zepslib = 1;
          break;
        }
        else if (!strcmp(optarg, "s2")) {
          zepslib2 = 1;
          zepslib = 1;
          break;
        }

      case 'a':
        if (!strcmp(optarg, "a")) {
          zetablib0 = zetablib1 = zetablib2 = 1;
          zetablib = 1;
          break;
        }
        else if (!strcmp(optarg, "a0")) {
          zetablib0 = 1;
          zetablib = 1;
          break;
        }
        else if (!strcmp(optarg, "a1")) {
          zetablib1 = 1;
          zetablib = 1;
          break;
        }
        else if (!strcmp(optarg, "a2")) {
          zetablib2 = 1;
          zetablib = 1;
          break;
        }

      case 'p':
        if (!strcmp(optarg, "p")) {
          zeplib0 = zeplib1 = zeplib2 = 1;
          zeplib = 1;
          break;
        }
        else if (!strcmp(optarg, "p0")) {
          zeplib0 = 1;
          zeplib = 1;
          break;
        }
        else if (!strcmp(optarg, "p1")) {
          zeplib1 = 1;
          zeplib = 1;
          break;
        }
        else if (!strcmp(optarg, "p2")) {
          zeplib2 = 1;
          zeplib = 1;
          break;
        }
      case 'f':
        if (!strcmp(optarg, "f")) {
          zeflib0 = 1;
	  zeflib1 = zeflib2 = 0;
          zeflib = 1;
          break;
        }
        else if (!strcmp(optarg, "f0")) {
          zeflib0 = 1;
          zeflib = 1;
          break;
        }
        else if (!strcmp(optarg, "f1")) {
          zeflib1 = 1;
          zeflib = 1;
          break;
        }
        else if (!strcmp(optarg, "f2")) {
          zeflib2 = 1;
          zeflib = 1;
          break;
        }
      default:
        Usage(argv[0]);
        exit(-1);
      }
      break;
      
    case 'h':
    case '?':
      Usage(argv[0]);
      exit(-1);
    }

  /* Lest loop */
  printf("Computing tests\n");
  printf("===============\n");
  TT = runtime();
  for(;;) {
    if (zen2lib) {
      /* Rg = $Z/2Z$ */
      
      if ((level[0] = zen2lib0)) {nb_rg=1;}
      if ((level[1] = zen2lib1)) {nb_rg=2;}
      if ((level[2] = zen2lib2)) {nb_rg=3;}
      zen2lib = 0;
      nl = 1;

      if (!(def = (char **)ZENMalloc(sizeof(char *)*NB_RG))) {
        fprintf(stderr, "Malloc error\n");
        exit(-1);
      }
      for(i = 0; i < NB_RG; i++) {
        if (ZEN_ZE2_DEF[i]) {
          if (!(def[i] = (char *) ZENMalloc(sizeof(char)*(1+strlen(ZEN_ZE2_DEF[i]))))) {
            fprintf(stderr, "Malloc error\n");
            exit(-1);
          }

#if 0
	  switch(i) {
	  case 0:
	    strcpy(def[i], ZEN_ZE2_DEF[i]);
	    break;
	  case 1:
	    strcpy(def[i], "(1)*X^10+(1)*X^3+(1)");
	    break;
	  case 2:
	    strcpy(def[i], "((0)*t^63+(1))*X^2+((0)*t^63+(1))*X+((0)*t^63+(1)*t^7)");
	    break; 
	  }
#else
          strcpy(def[i], ZEN_ZE2_DEF[i]);
#endif
        }
        else {
          def[i] = NULL;
        }
      }       
    }
    else if (zepslib) {
      /* Rg = $Z/pZ$, $p < 2^{BN_DIGIT_SIZE}$ */
      
      if ((level[0] = zepslib0)) {nb_rg=1;}
      if ((level[1] = zepslib1)) {nb_rg=2;}
      if ((level[2] = zepslib2)) {nb_rg=3;}
      zepslib = 0;
      nl = 1;

      if (!(def = (char **)ZENMalloc(sizeof(char *)*NB_RG))) {
        fprintf(stderr, "Malloc error\n");
        exit(-1);
      }
      for(i = 0; i < NB_RG; i++) {
        if (ZEN_ZEPS_DEF[i]) {
          if (!(def[i] = (char *) ZENMalloc(sizeof(char)*(1+strlen(ZEN_ZEPS_DEF[i]))))) {
            fprintf(stderr, "Malloc error\n");
            exit(-1);
          }
          strcpy(def[i], ZEN_ZEPS_DEF[i]);
        }
        else {
          def[i] = NULL;
        }
      }       
    }
    else if (zeplib) {
      /* Rg = $Z/pZ$, $p < 2^{BN_DIGIT_SIZE}$ */
      
      if ((level[0] = zeplib0)) {nb_rg=1;}
      if ((level[1] = zeplib1)) {nb_rg=2;}
      if ((level[2] = zeplib2)) {nb_rg=3;}
      zeplib = 0;
      nl = 1;

      if (!(def = (char **)ZENMalloc(sizeof(char *)*NB_RG))) {
        fprintf(stderr, "Malloc error\n");
        exit(-1);
      }
      for(i = 0; i < NB_RG; i++) {
        if (ZEN_ZEP_DEF[i]) {
          if (!(def[i] = (char *) ZENMalloc(sizeof(char)*(1+strlen(ZEN_ZEP_DEF[i]))))) {
            fprintf(stderr, "Malloc error\n");
            exit(-1);
          }
          strcpy(def[i], ZEN_ZEP_DEF[i]);
        }
        else {
          def[i] = NULL;
        }
      }       
    }
    else if (zeflib) {
      /* Rg = $Z/pZ$, $p < 2^{BN_DIGIT_SIZE}$ */

#if APPROXIMATE_RESULTS
      fprintf(stderr,"WARNING: APPROXIMATE_RESULTS may cause error in zentest \nbecause equality cannot be strictly defined.\n APPROXIMATE_LEVEL was set to %d\n",APPROXIMATE_LEVEL);
#endif
      if(zeflib1 || zeflib2)
	fprintf(stderr,"WARNING: Zef may cause errors when used in extension towers.\n");
      
      if ((level[0] = zeflib0)) {nb_rg=1;}
      if ((level[1] = zeflib1)) {nb_rg=2;}
      if ((level[2] = zeflib2)) {nb_rg=3;}
      zeflib = 0;
      nl = 1;

      if (!(def = (char **)ZENMalloc(sizeof(char *)*NB_RG))) {
        fprintf(stderr, "Malloc error\n");
        exit(-1);
      }
      for(i = 0; i < NB_RG; i++) {
        if (ZEN_ZEF_DEF[i]) {
          if (!(def[i] = (char *) ZENMalloc(sizeof(char)*(1+strlen(ZEN_ZEF_DEF[i]))))) {
            fprintf(stderr, "Malloc error\n");
            exit(-1);
          }
          strcpy(def[i], ZEN_ZEF_DEF[i]);
        }
        else {
          def[i] = NULL;
        }
      }       
    }
    else if (zetablib) {
      /* Rg = $Z/pZ$, $p < 2^{BN_DIGIT_SIZE}$ */
      
      if ((level[0] = zetablib0)) {nb_rg=1;}
      if ((level[1] = zetablib1)) {nb_rg=2;}
      if ((level[2] = zetablib2)) {nb_rg=3;}
      zetablib = 0;
      nl = 1;

      if (!(def = (char **)ZENMalloc(sizeof(char *)*NB_RG))) {
        fprintf(stderr, "Malloc error\n");
        exit(-1);
      }
      for(i = 0; i < NB_RG; i++) {
        if (ZEN_ZETAB_DEF[i]) {
          if (!(def[i] = (char *) ZENMalloc(sizeof(char)*(1+strlen(ZEN_ZETAB_DEF[i]))))) {
            fprintf(stderr, "Malloc error\n");
            exit(-1);
          }
          strcpy(def[i], ZEN_ZETAB_DEF[i]);
        }
        else {
          def[i] = NULL;
        }
      }       
    }
    else break;

    for(rg = 0; rg < nb_rg; rg++) {
      switch(rg)
        {
        case 0:
          /* Initialization */
          if (ZBNReadFromString(&n, &nl, def[0], 10) != strlen(def[0])) {
            ZENError();
            fprintf(stderr, "Error while reading a bignum\n");
            exit(-1);
          }

          printf("\nInitializing the ring R0 = Z/");
          if (ZBNPrint(n, nl)) {
            ZENError();
            fprintf(stderr, "Error while printing a bignum\n");
            exit(-1);
          }
          printf("Z\n");
          tt = runtime();
          if (ZENBaseRingAlloc(Rg[0], n, nl)) {
            ZENError();
            fprintf(stderr, "Error while initializing this ring\n");
            exit(-1);
          }
          ZBNF(n);
          printf("Initialization of the ring R0 done in %.2f s\n",
                 (runtime()-tt)/1000.);

	  break;

        default :
	  /* Put all the precomputation flags in prc */
	  ZENPrcSetAll(prc);
            
          /* Initialization of the extension ring */
          if (def[rg]) {
            printf("\nInitializing the ring R%d = R%d", rg, rg-1);
            printf("[X]/(P(X)) with\nP(X) = %s\n", def[rg]);
            tt = runtime();

            if(ZENPolyReadFromString(P1, def[rg], 10, Rg[rg-1]) < 0) {
              ZENError();
              fprintf(stderr, "Error while reading the definition polynomial\n");
              exit(-1);
            }
            
            if (ZENExtRingAlloc(Rg[rg], P1, Rg[rg-1])) {
              ZENError();
              fprintf(stderr, "Error while initilizing the ring\n");
              exit(-1);
            }
          }
          else {
            printf("\nInitializing the ring R%d = R%d", rg, rg-1);
            printf("[X]/(P(X)) with degree(P(X)) = %d :\n", DEG_ZEN_EXT);
            tt = runtime();
            if (!(P1 = ZENPolyCreate(DEG_ZEN_EXT, Rg[rg-1]))) {
              ZENError();
              fprintf(stderr, "Error while allocation P(X)\n");
              exit(-1);
            }
#if FINITE_FIELD
            ZENPolySetRandomIrreducible(P1, DEG_ZEN_EXT, Rg[rg-1]);
#else
            ZENPolySetRandom(P1, DEG_ZEN_EXT, Rg[rg-1]);
#endif
            ZENPolyPrintToFile(stdout,P1,10,Rg[rg-1]);
            printf("\n");
            
            for(;;) {
              if (!ZENExtRingAlloc(Rg[rg], P1, Rg[rg-1])) 
                break;
              
              if (ZENErrErrno != ZEN_NONMONIC) {
                ZENError();
                fprintf(stderr, "Error while initializing this ring\n");
                exit(-1);
              }
              ZENPolySetRandom(P1, DEG_ZEN_EXT, Rg[rg-1]);
            }
          }
          ZENPolyFree(P1, Rg[rg-1]);
          printf("Initialization of the ring R%d done in %.2f s\n", rg, (runtime()-tt)/1000.);          
        }

      for(first = 0;; first++) {
	
	if (!first) {

	  /* Clone */
	  ZENClnSetAll(cln);

	  printf("Computing a clone of the ring R%d\n", rg);
	  tt = runtime();
	  RgClone = ZENRingClone(Rg[rg], cln);
	  printf("A clone of the ring R%d done in %.2f s\n",
		 rg, (runtime()-tt)/1000.);

	}
	else if (first == 1) {
	  if (ZENRingArith(Rg[rg]) == ZENRingArith(RgClone)) {
	    printf("The clone is the same ring as R%d, I don't test it\n",
		   rg);

	    /* Freeing the clone */
	    printf("\nClosing the clone of ring R%d\n", rg);
	    tt = runtime();
	    if (ZENRingClose(RgClone)) {
	      fprintf(stderr, "Error while closing this ring\n");
	      ZENError();
	      exit(-1);
	    }
	    printf("This clone of ring was succesfully closed in %.2f s\n", (runtime()-tt)/1000.);
	    break;
	  }
	  oldrg = Rg[rg];
	  Rg[rg] = RgClone;
	}
	else if (first == 2) {

	  /* Freeing the clone */
	  printf("\nClosing the clone of ring R%d\n", rg);
	  tt = runtime();
	  if (ZENRingClose(RgClone)) {
	    fprintf(stderr, "Error while closing this ring\n");
	    ZENError();
	    exit(-1);
	  }
	  printf("This clone of ring was succesfully closed in %.2f s\n", (runtime()-tt)/1000.);
	  Rg[rg] = oldrg;
	  break;
	}

	if (level[rg]) {

	  /* Precomputations */
	  ZENPrcSetAll(prc);
	  ZENPrcSet(prc, ZENPRC_FINITE_FIELD);
	  printf("Precomputations for the ring R%d\n", rg);
	  tt = runtime();
	  ZENRingAddPrc(Rg[rg], prc);
	  printf("Precomputations of the ring R%d done in %.2f s\n",
		 rg, (runtime()-tt)/1000.);
	
	  if (ZENTestPolyFactor(Rg[rg], NB_TEST_POLY_FACTOR, DEG_POLY)) {
	    fprintf(stderr, "Error while testing ZENPolyRoots()\n");
	    ZENError();
	    exit(-1);
	  }
	  if (ZENTestPolyRoots(Rg[rg], NB_TEST_POLY_ROOTS, DEG_POLY)) {
	    fprintf(stderr, "Error while testing ZENPolyRoots()\n");
	    ZENError();
	    exit(-1);
	  }
	  if (ZENTestPolyIsNotIrreducible(Rg[rg], NB_TEST_POLY_IRREDUCIBLE, DEG_POLY)) {
	    fprintf(stderr, "Error while testing ZENPolyIsNotIrreducible()\n");
	    ZENError();
	    exit(-1);
	  }
	}
      }
    }
      
    for(rg--; rg >= 0; rg--) {
      
      /* Freeing the extension rings */
      printf("\nClosing the ring R%d\n", rg);
      tt = runtime();
      if (ZENRingClose(Rg[rg])) {
        fprintf(stderr, "Error while closing this ring\n");
        ZENError();
        exit(-1);
      }
      printf("This ring was succesfully closed in %.2f s\n", (runtime()-tt)/1000.);
    }
    
    /* free constants */
    for(i = 0; i < NB_RG; i++) {
      if (def[i]) {
        ZENFree(def[i]);
      }
    }
    ZENFree(def);
  }
  printf("\nAll the tests correctly performed in %.2f s\n", (runtime()-TT)/1000.);

  
#if defined(i486icc) & defined(__DEBUG_ALLOC__)
  _dump_allocated(0);
#endif

  exit(0);
}

void Usage(arg)
     char *arg;
{
  printf("Usage: %s [-l(t|s)[val]] [-help]\n", arg);
  printf("This program perform tests of the library zenfact\n");
  printf("If no option is given, all the functions are tested\n");
  printf("-l <lib>: To perform tests only on function of the ring <lib>.\n");
  printf("<lib> can be,\n");
  printf("     t : for the subring Z/2Z.\n");
  printf("     a : for the subring Z/pZ, p very small.\n");
  printf("     s : for the subring Z/pZ, p small.\n");
  printf("     p : for the subring Z/pZ.\n");
  printf("     f : for Q.\n");
  printf(" [val] can be the level of extensions, ie 0, 1 or 2.\n");
}
