#include <stdio.h>
#include "zen.h"
#include "zenfact.h"

#ifdef ___
static int __ZENMillerRabin ___((ZENRing,BigNumDigit,ZENElt,BigNum,BigNumLength,int));
#endif
static int __ZENMillerRabin(Rg, e, elt, k, kl,q)
     ZENRing Rg;
     BigNumDigit e;
     ZENElt elt;
     BigNum k;    
     BigNumLength kl;
     int q;
{
  int i;
  BigNumDigit a[1];
  
  /* a^k */
  a[0]=(BigNumDigit)e;
  ZENZToElt(elt,a,1, Rg);
  if(ZENEltExp(elt, k, kl, elt, Rg)) {
    return(ZERR);
  }
	 
  if (ZENEltIsOne(elt, Rg)) {
    return(1);
  }
  ZENEltNegate(elt, elt, Rg);
  if (ZENEltIsOne(elt, Rg)) {
    return(1);
  }

  for(i = 1; i < q; i++) {
    
    /* a^{2^ik} */
    ZENEltSquare(elt, elt, Rg);
    ZENEltNegate(elt, elt, Rg);
	
    if (ZENEltIsOne(elt, Rg)) {
      return(1);
    }
  }
  return(0);
}

#if 0 
/*
   \TEX{
   \title{Miller Rabin test}
   \input{A finite ring |Rg| and a base |e|.}
   \output{|ZENERR| if an error occured, 1 if |Rg| is a strong finite field
   in base |e|, 0 otherwise.}
   \note{The first integer  which is a strong pseudo prime
   for |e| = 2, 3, 5, 7, and 11, but not a prime is larger than $25
   10^{12}$ }
   } 
   */
int ZENMillerRabin(Rg, e)
     ZENRing Rg;
     BigNumDigit e;{}
#endif
int _ZENMillerRabin(Rg, e)
     ZENRing Rg;
     BigNumDigit e;
{
  BigNum k;
  BigNumLength kl;
  int q,ret;
  ZENElt elt;

  if((ZENRingQl(Rg)==1)		/* q <= e : I can't say anything... */
     &&(ZBNDigit(ZENRingQ(Rg),0)<=e))
    return(1);

#define PRECALCULS \
  if (!(k = ZBNC(ZENRingPl(Rg)))) { \
    ZENSetError(ZEN_LIB_ZEG, ZEN_MALLOC, ZEN_ALLOC); \
    return(ZERR);  } \
  ZBNAssign(k, ZENRingP1(Rg), ZENRingPl(Rg)); \
  if (ZBNIsOne(k, ZENRingPl(Rg))) { \
    ZBNF(k); \
    return(1); } \
  if (ZBNIsZero(k, ZENRingPl(Rg)) || (ZBNDigit(k, 0) & 1)) { \
    ZBNF(k); \
    return(0);  } \
  q=0; \
  do {   \
    ZBNShiftRight(k, ZENRingPl(Rg), 1); q++; \
  } while(!(ZBNDigit(k, 0) & 1));   \
  kl = ZBNNumDigits(k, ZENRingPl(Rg)); \
  if (ZENEltCreate(elt, Rg)) { \
    ZENSetError(ZEN_LIB_ZEG, ZEN_MALLOC, ZEN_ALLOC); \
    ZBNF(k); \
    return(ZERR);  }
  
  PRECALCULS;
  
  ret=__ZENMillerRabin(Rg,e,elt,k,kl,q);
  ZENEltFree(elt, Rg);
  ZBNF(k);
  return(ret);
}

#ifdef ___
static int __ZENIsPrime ___((ZENRing));
#endif
static int __ZENIsPrime(Rg)
  ZENRing Rg;
{
  int i,lim,q;
  ZENElt elt;
  BigNum k;
  BigNumLength kl;

  /*
   \TEX{
  The number of bases used in the Miller-Rabin test can be adjusted in the file |psprime.c|. 
   } 
   \CiteC */
#ifndef NB_BASIS_MILLER_RABIN
# define NB_BASIS_MILLER_RABIN 39 /* maximal number is 39 */
#endif
  int base[40];
  base[0]=2; base[1]=3; base[2]=5; base[3]=7; base[4]=11;
  base[5]=13;base[6]=17;base[7]=19;base[8]=23;base[9]=29;
  base[10]=31;base[11]=37;base[12]=41;base[13]=43;base[14]=47;
  base[15]=53;base[16]=59;base[17]=61;base[18]=67;base[19]=71;
  base[20]=73; base[21]=79; base[22]=83; base[23]=89; base[24]=97;
  base[25]=101;base[26]=103;base[27]=107;base[28]=109;base[29]=113;
  base[30]=127;base[31]=131;base[32]=137;base[33]=139;base[34]=149;
  base[35]=151;base[36]=157;base[37]=163;base[38]=167;base[39]=173;
 /* \EndCiteC */
  PRECALCULS;

  if(ZENRingQl(Rg)==1)
    {
      for(i=0,lim = -1;i<=NB_BASIS_MILLER_RABIN;i++)
	if(ZBNDigit(ZENRingQ(Rg),0)>(BigNumDigit)base[i])
	  lim=i;
	else
	  break;
    }
  else
    lim=NB_BASIS_MILLER_RABIN;

  for (i=0;i<=lim;i++)
    switch(__ZENMillerRabin(Rg,(BigNumDigit)base[i],elt,k,kl,q))
      {
      case 0:
	ZENEltFree(elt, Rg);
	ZBNF(k);	
	return(0);
      case 1:
	break;
      default:
	ZENEltFree(elt, Rg);
	ZBNF(k);
	return(ZERR);
      }
  ZENEltFree(elt, Rg);
  ZBNF(k);
  return(1);  
}
#if 0
/* \TEX{
   \title{Is it a pseudo finite field.}
   \input{A finite ring |Rg|.}
   \output{|ZENERR| if an error occurred, 1 if |Rg| is a pseudo finite
   field, 0 otherwise.}
   \note{This procedure test if the definition ring is a finite field
   and then, if necessary, if the definition polynomial is
   irreducible. When |Rg| is a prime field, this procedure test if
   it is a field with the Miller Rabin test applied to several basis. With
   2, 3, 5, 7, 11 and 13, the answer is exact for all $n<10^{12}$}
   } */
int ZENRingIsAPseudoFF(Rg)
     ZENRing Rg;{}
#endif
int _ZENRingIsAPseudoFF(Rg)
     ZENRing Rg;
{
  int ret;

  if (ZENRingDef(Rg)) {

    /* Extension case */
    if ((ret = ZENRingIsAPseudoFF(ZENRingDef(Rg))) != 1) {
      return(ret);
    }
    switch (ZENPolyIsNotIrreducible(ZENRingPol(Rg), ZENRingDef(Rg))) {
    case ZENERR:
      ret = ZERR;
      break;
    case ZENPOLY_IS_IRREDUCIBLE:
      ret = 1;
      break;
    default:
      ret = 0;
      break;
    }
  }
  else {
    ret=__ZENIsPrime(Rg);
  }
  return(ret);
}

/*
   \TEX{
   \title{Is an integer prime}
   \input{A |BigNum| and its length.}
   \output{|ZENERR| if an error occurs, 1 if the integer is pseudo-prime, 0 otherwise}
   \note{The test uses |ZENRingIsAPseudoFF|}
   } 
   */
int ZBNIsPrime(n,nl)
     BigNum n;
     BigNumLength nl;
{
  ZENRing R;
  int ret;

  nl=ZBNNumDigits(n,nl);
  if((nl==1)&&(((*n)|(BigNumDigit)1) == (BigNumDigit)1))	/* n==0 or n==1 */
    return(0);
  if(ZENBaseRingAlloc(R,n,nl))
    return(ZERR);
  ret=__ZENIsPrime(R);
  ZENRingClose(R);
  return(ret);
}

/*
   \TEX{
   \title{Find the prime immediatly following}
   \input{A pointer to an allocated |BigNum| and a pointer to its length.}
   \output{|ZENERR| if an error occurs, the number of steps performed otherwise}
   \side{The |BigNum| is modified accordingly. If a reallocation is needed, it is performed after freeing the initial |BigNum|.}
   \note{The test uses |ZBNIsPrime|. If the number is already prime, it is unchanged.}
   } 
   */
int ZBNNextPrime(p_n,p_nl)
     BigNum *p_n;
     BigNumLength *p_nl;
{
  int ret=0;
  BigNumLength nl;

  if((ZBNNumDigits(*p_n,*p_nl)==1)
     &&((*p_n)[0]<=(BigNumDigit)2))	/* cas de 0,1,2 */
    {
      ret=2-((*p_n)[0]);
      (*p_n)[0]=(BigNumDigit)2;
      return(ret);
    }
  
  for(;;)
    if((*p_n)[0]&1)
      switch(ZBNIsPrime(*p_n,*p_nl))
	{
	case ZENERR:
	  return(ZERR);
	case 1:
	  return(ret);
	case 0:
	  ret++;
	  if(ZBNAddCarry(*p_n,*p_nl,1))
	    {
	      ZBNF(*p_n);
	      *p_n=ZBNC(*p_nl+1);
	      ZBNSetToZero(*(p_n+1),*p_nl-1);
	      (*p_n)[*p_nl]=(*p_n)[0]=1;		/* A power of 2 is not prime ! */
	      (*p_nl)++;	      
	    }
	  break;
	default:
	  return(ZERR);
	}
    else
      {
	ret++;
	(*p_n)[0] |= 1;
      }  
}
