#include <stdio.h>
#include "zen.h"
#include "zec.h"
#include "zenfact.h"


#if 0
/*
  \TEX{
  \title{Generating a finite field}
  \input{A |ZENRing| to build, a number |(q,ql)| and the degree |d| of the extension.}
  \output{|ZENErr| if an error occured, 0 otherwise.}
  \side{The |ZENRing| $$R=\quotient{\quotient{\Z}{p\Z}[X]}{P(X)}$$ is built using an irreducible polynomial of degree |d| generated by |ZENPolySetGoodIrreducible|. The prime $p$ is the smallest prime greater or equal than |(q,ql)|.}
  \note{No precomputation is done. If |d|$<2$ the built field is $R=\quotient{\Z}{p\Z}$. The base ring is cloned, in order to improve efficiency, using |ZENClnSetAll| and |ZENRingClone|. The extension is cloned in the same way. The obtained structure can be freed using |ZENRingFullClose|.}
  }
   */
int ZENFieldAlloc(R,q,ql,d)
     ZENRing R;
     BigNum q;
     BigNumLength ql;
     int d;{}
#endif
ZENRing _ZENFieldCreate(q,ql,d)
     BigNum q;
     BigNumLength ql;
     int d;
{
  BigNum p;
  BigNumLength pl;
  ZENRing Rg,Cl,Ex,ret;
  ZENPoly P;
  ZENCln cln;
  
  pl=ql;
  p=ZBNC(pl);
  if(p==NULL)
    {
      ZENSetError(ZEN_LIB_ZEG, ZEN_INIT, ZEN_ALLOC);
      return(ZNULL);
    }
  ZBNAssign(p,q,ql);  
  if(ZBNNextPrime(&p,&pl)<0) {
    ZBNF(p);
    return(ZNULL);
  }
  
  Rg=ZENBaseRingCreate(p,pl);
  ZBNF(p);
  if(Rg==NULL)
    return(ZNULL);
  
  ZENClnSetAll(cln);
  Cl=ZENRingClone(Rg,cln);
  if(Cl==NULL)
    {
      ZENRingClose(Rg);
      return(ZNULL);
    }
  
  if(d<2)
    return(Cl);

  if(ZENPolyAlloc(P,d,Cl)<0)
    {
      ZENRingClose(Cl);
      ZENRingClose(Rg);
      return(ZNULL);
    }
  if(ZENPolySetGoodIrreducible(P,d,Cl)<0)
    {
      ZENPolyFree(P,Cl);
      ZENRingClose(Cl);
      ZENRingClose(Rg);
      return(ZNULL);
    }
  Ex=ZENExtRingCreate(P,Cl);
  ZENPolyFree(P,Cl);
  if(Ex==NULL)
    {
      ZENRingClose(Cl);
      ZENRingClose(Rg);
      return(ZNULL);
    }      
  ret=ZENRingClone(Ex,cln);
  if(ret==NULL)
    {
      ZENRingClose(Ex);
      ZENRingClose(Cl);
      ZENRingClose(Rg);
      return(ZNULL);
    }
  return(ret);
}

/**********************************************************************/
static char *Read(fin)
     FILE *fin;
{
  char *str,*str2;
  char line[1024];
 
  str = (char *)ZENMalloc(2*sizeof(char));
  if(str==NULL) {
    ZENSetError(ZEN_LIB_ZEG, ZEN_INIT, ZEN_ALLOC);
    return(NULL);
  }
  sprintf(str,"\\");
  for(;;)
    {
      if(fscanf(fin, "%s", line)!=1) {
	ZENSetError(ZEN_LIB_ZEG, ZEN_INIT, ZEN_FSCANF);
	return(NULL);
      }
	
      str2=str;
      str=(char *)ZENMalloc((strlen(str)+strlen(line))*sizeof(char));
      if(str==NULL) {
	ZENSetError(ZEN_LIB_ZEG, ZEN_INIT, ZEN_ALLOC);
	ZENFree(str2);
	return(NULL);
      }
      if(sprintf(str,"%s",str2)<0) {
	ZENSetError(ZEN_LIB_ZEG, ZEN_INIT, ZEN_SPRINTF);
	ZENFree(str);
	ZENFree(str2);
	return(NULL);
      }
      ZENFree(str2);
      if(sprintf(str+strlen(str)-1,"%s",line)<0) {
	ZENSetError(ZEN_LIB_ZEG, ZEN_INIT, ZEN_SPRINTF);
	ZENFree(str);
	return(NULL);
      }
      if(str[strlen(str)-1]!='\\')
	break;
    }
  return(str);
}

#if 0
/*
   \TEX{
   \title{Generating interactively a finite ring}
   \input{Two file descriptors that can be |stdin| and |stdout|.}
   \output{|ZENNULL| if an unrecoverable error occured or the newly defined |ZENRing|.}
   \side{All the possible clonings are performed, but no precomputation is done.}
   \note{If |OUT| is set to |NULL|, the interactive questions are not output.} 
   } 
   */
ZENRing ZENRingDefine(IN,OUT)
     FILE *IN,*OUT;{}
#endif
ZENRing _ZENRingDefine(IN,OUT)
     FILE *IN,*OUT;
{
#define NBRGALLOC 5
  BigNum n;
  BigNumLength nl;
  char *str;
  double tt;
  int rg,alloc,ext,i;
  ZENPoly P;
  ZENCln cln;
  ZENRing Rg,Ex,*Rgopt;

  alloc=NBRGALLOC;
  Ex=ZENNULL;
  Rgopt=(ZENRing *)(ZENMalloc(alloc*sizeof(ZENRing)));
  if(Rgopt==NULL)
    {
      ZENSetError(ZEN_LIB_ZEG, ZEN_INIT, ZEN_ALLOC);
      return(ZNULL);
    }

  for(rg=ext=0;;rg++)
    {
      int CRT;
      
      if(OUT) {
	fprintf(OUT,"\n***********\nEnd the line by '\\' if you need several lines.\n");
	if(rg) {
	  fprintf(OUT,"Enter 1 to output the ring\n");
	  fprintf(OUT,"In order to use Chinese Remainder Theorem, begin your input with 'C'\n");
	}
	if(!ext)
	  {
	    fprintf(OUT,"Enter the integer n which defines Z/nZ:\n");
	    fprintf(OUT,"options: +<number> for using the smallest prime greater than <number>;\n");
	  }
	if(rg || ext)
	  {
	    fprintf(OUT,"In order to build an extension, enter a polynomial.\n");
	    fprintf(OUT,"options: ^<degree> for making an extension of given degree.\n");
	  }
      }
      str=Read(IN);
      if(str==NULL)
	{
	  for(i=1;i<rg;i++)
	    ZENRingCloneClose(Rgopt[i]);
	  ZENRingFullClose(Rgopt[0]);
	  ZENFree(Rgopt);
	  return(ZNULL);
	}
      if((str[0]=='C')||(str[0]=='c')) /* If Chinese Remainder Theorem */
	{
	  if(rg==0)
	    {
	      ZENSetError(ZEN_LIB_ZEG, ZEN_INIT, ZEN_SYNTAX);
	      if(Ex)
		ZENRingFullClose(Ex);
	      ZENFree(str);
	      ZENFree(Rgopt);
	      return(ZNULL);
	    }
	  CRT=1;
	  str++;
	}
      else
	{
	  CRT=0;
	  if(rg>=2)		/* else we compute the CRT ring xith the previously
				   defined rings */
	    {
	      if(ZENRingChinese(Rg,rg,Rgopt)<0) {
		for(i=1;i<rg;i++)
		  ZENRingCloneClose(Rgopt[i]);
		ZENRingFullClose(Rgopt[0]);
		ZENFree(Rgopt);
		return(ZNULL);
	      }
	      Rgopt[0]=Rg;
	    }
	  if(rg)
	    ext=1;
	  rg=0;
	  Ex=Rgopt[0];
	}
      if((str[0]=='(')||(str[0]=='^')) /* if polynomial */
	{
	  ext=1;
	  if(str[0]=='(')
	    {
	      if(ZENPolyReadFromString(P, str, 10, Ex) < 0)
		{
		  if(CRT)
		    {
		      for(i=1;i<rg;i++)
			ZENRingCloneClose(Rgopt[i]);
		      str--;
		    }
		  ZENRingFullClose(Ex);
		  ZENFree(str);
		  ZENFree(Rgopt);
		  return(ZNULL);
		}
	      if(CRT)
		str--;
	      ZENFree(str);
	    }
	  else			/* if polynomial degree given */
	    {
	      int deg;
	  
	      if(sscanf(str+1,"%d",&deg)!=1)
		{
		  ZENSetError(ZEN_LIB_ZEG, ZEN_INIT, ZEN_SSCANF);
		  if(CRT)
		    {
		      for(i=1;i<rg;i++)
			ZENRingCloneClose(Rgopt[i]);
		      str--;
		    }
		  ZENRingFullClose(Ex);
		  ZENFree(str);
		  ZENFree(Rgopt);
		  return(ZNULL);
		}
	      if(CRT)
		str--;
	      ZENFree(str);
	      if(ZENPolyAlloc(P,deg,Ex)<0)
		{
		  if(CRT)
		    {
		      for(i=1;i<rg;i++)
			ZENRingCloneClose(Rgopt[i]);
		    }
		  ZENRingFullClose(Ex);
		  ZENFree(Rgopt);
		  return(ZNULL);
		}
	      tt=runtime();
	      if(ZENPolySetGoodIrreducible(P,deg,Ex)<0)
		{
		  ZENPolyFree(P,Ex);
		  if(CRT)
		    {
		      for(i=1;i<rg;i++)
			ZENRingCloneClose(Rgopt[i]);
		    }
		  ZENRingFullClose(Ex);
		  ZENFree(Rgopt);
		  return(ZNULL);
		}
	      if(OUT) {
		fprintf(OUT,"Good irreducible polynomial found in %.2f s\n",(runtime()-tt)/1000.);
	      }	      
	    }
	  if(OUT) {
	    fprintf(OUT,"Making extension modulo\n");
	    ZENPolyPrintToFile(OUT,P,10,Ex);
	    fprintf(OUT,"\n");
	  }
	  tt=runtime();
	  if(ZENExtRingAlloc(Rg, P, Ex)<0)
	    {
	      ZENPolyFree(P,Ex);
	      if(CRT)
		{
		  for(i=1;i<rg;i++)
		    ZENRingCloneClose(Rgopt[i]);
		}
	      ZENRingFullClose(Ex);
	      ZENFree(Rgopt);
	      return(ZNULL);
	    }
	  ZENPolyFree(P,Ex);
	  if(OUT) {
	    fprintf(OUT,"Extension ring R%d done in %.2f s\n",rg,(runtime()-tt)/1000.);
	  }
	}
      else			/* It's a 1 or a BigNum */
	{
	  if(str[0]== '+')	/* Computing a prime */
	    {
	      if(ext) {		/* If already in an extension non sense */
		ZENSetError(ZEN_LIB_ZEG, ZEN_INIT, ZEN_SYNTAX);
		if(CRT)
		  {
		    for(i=1;i<rg;i++)
		      ZENRingCloneClose(Rgopt[i]);
		    ZENRingFullClose(Rgopt[0]);
		    str--;
		  }
		else
		  ZENRingFullClose(Ex);
		ZENFree(str);
		ZENFree(Rgopt);
		return(ZNULL);
	      }
	      if (ZBNReadFromString(&n, &nl, str+1, 10) < 0) {
		ZENError();
		if(CRT)
		  {
		    for(i=1;i<rg;i++)
		      ZENRingCloneClose(Rgopt[i]);
		    ZENRingFullClose(Rgopt[0]);
		    str--;
		  }
		ZENFree(str);
		ZENFree(Rgopt);
		return(ZNULL);
	      }
	      if(CRT)
		str--;
	      ZENFree(str);
	      tt=runtime();
	      if (ZBNNextPrime(&n,&nl)<0) {
		ZENError();
		ZBNF(n);
		if(CRT)
		  {
		    for(i=1;i<rg;i++)
		      ZENRingCloneClose(Rgopt[i]);
		    ZENRingFullClose(Rgopt[0]);
		  }
		ZENFree(Rgopt);
		return(ZNULL);
	      }
	      if(OUT) {
		fprintf(OUT,"Prime characteristic computed in %.2f s.\n",(runtime()-tt)/1000.);
	      }
	    }
	  else if(isdigit(str[0])) /* Prime given */
	    {
	      if (ZBNReadFromString(&n, &nl, str, 10) < 0) {
		ZENError();
		if(CRT)
		  {
		    for(i=1;i<rg;i++)
		      ZENRingCloneClose(Rgopt[i]);
		    ZENRingFullClose(Rgopt[0]);
		    str--;
		  }
		ZENFree(str);
		ZENFree(Rgopt);
		return(ZNULL);
	      }
	      if(CRT)
		str--;
	      ZENFree(str);
	    }
	  else			/* This is a syntax error */
	    {
	      ZENSetError(ZEN_LIB_ZEG, ZEN_INIT, ZEN_SYNTAX);
	      if(CRT)
		{
		  for(i=1;i<rg;i++)
		    ZENRingCloneClose(Rgopt[i]);
		  ZENRingFullClose(Rgopt[0]);
		  str--;
		}
	      ZENFree(str);
	      ZENFree(Rgopt);
	      return(ZNULL);
	    }

	  if ((ZBNDigit(n, 0) == 1) && (nl == 1)) { /* Now testing if it is 1 */
	    ZBNF(n);
	    if(CRT)		/* if CRT syntax error */
	      {
		for(i=1;i<rg;i++)
		  ZENRingCloneClose(Rgopt[i]);
		ZENRingFullClose(Rgopt[0]);
		ZENFree(Rgopt);
		ZENSetError(ZEN_LIB_ZEG, ZEN_INIT, ZEN_SYNTAX);
		return(ZNULL);
	      }
	    ZENFree(Rgopt);
	    return(Ex);		/* returning the result */
	  }
	  if(ext) {
	    ZENSetError(ZEN_LIB_ZEG, ZEN_INIT, ZEN_SYNTAX);
	    if(CRT)
	      {
		for(i=1;i<rg;i++)
		  ZENRingCloneClose(Rgopt[i]);
		ZENRingFullClose(Rgopt[0]);
	      }
	    else
	      ZENRingFullClose(Ex);
	    ZENFree(Rgopt);
	    return(ZNULL);
	  }
	    
	  if(OUT) {
	    fprintf(OUT,"\nInitializing the ring R%d = Z/",rg);
	    ZBNPrintToFile(OUT,n, nl,10);
	    fprintf(OUT,"Z\n");
	  }
	  tt = runtime();
	  if (!(Rg = ZENBaseRingCreate(n, nl))) {
	    ZENError();
	    ZBNF(n);
	    if(CRT)
	      {
		for(i=1;i<rg;i++)
		  ZENRingCloneClose(Rgopt[i]);
		ZENRingFullClose(Rgopt[0]);
	      }
	    ZENFree(Rgopt);
	    return(ZNULL);
	  }
	  ZBNF(n);
	  if(OUT) {
	    fprintf(OUT,"Initialization of the ring R%d done in %.2f s\n",rg,(runtime()-tt)/1000.);
	  }
	}
	      
      tt = runtime();		/* Now making the clone */
      ZENClnSetAll(cln);
      if(rg>=alloc)		/* realloc if needed */
	{
	  alloc += NBRGALLOC;
	  Rgopt=(ZENRing *)(ZENRealloc(Rgopt,alloc*sizeof(ZENRing)));
	  if(Rgopt==NULL)
	    {
	      if(CRT)
		{
		  for(i=0;i<rg;i++)
		    ZENRingCloneClose(Rgopt[i]);
		}
	      ZENRingFullClose(Rg);
	      ZENSetError(ZEN_LIB_ZEG, ZEN_INIT, ZEN_ALLOC);
	      ZENFree(Rgopt);
	      return(ZNULL);
	    }
	}
      if (!(Rgopt[rg] = ZENRingClone(Rg, cln))) {
	ZENError();
	if(CRT)
	  {
	    for(i=0;i<rg;i++)
	      ZENRingCloneClose(Rgopt[i]);
	  }
	ZENRingFullClose(Rg);
	ZENFree(Rgopt);
	return(ZNULL);
      }
      if(OUT) {
	fprintf(OUT,
		"Cloning of the ring Rg%d done in %.2f s\n",rg,(runtime()-tt)/1000.);
      }
    } 
}

#if 0
/* \TEX{
   \title{Precomputing.}
   \input{A finite ring |Rg| and precomputations |Prc|.} 
   \output{|ZENERR| if an error occured, |ZEN_NO_INVERSE| if a factor of a
   modulo was discovered, |ZEN_HAS_INVERSE| otherwise.}
   \side{Precomputations asked in the flags of |Prc| are done in order
     to speed up the corresponding procedures. |ZENRingFact(Rg)| is
     filled with a factor of a modulo if |ZEN_NO_INVERSE| is
     returned. These precomputations are done recursively for all the
     underlying |ZENRing|s. If a clone is encountered, precomputations
     are added to the clone, not to the original ring, because its
     operations are not used.}
   \note{Precomputations can take time\ldots This function can be used
   for instance on the output of a |_ZENRingDefine| call.}
} */
int ZENRingRecursiveAddPrc(R, prc)
     ZENRing R;
     ZENPrc prc;{}
#endif
extern int _ZENRingRecursiveAddPrc(R, prc)
     ZENRing R;
     ZENPrc prc;
{
  switch(ZENRingArith(R))
    {
    case ZEN_LIB_ZEC:
      {
	int i;
	ZENRing *PR;

	PR=ZecRingSubrings(R);
	i=ZecRingNbS(R);	
	for(i--;i>=1;i--)
	  {
	    if(ZENRingAddPrc(PR[i],prc))
	      return(ZERR);
	  }
	if(ZENRingRecursiveAddPrc(PR[0],prc))
	  return(ZERR);
	break;
      }
    case ZEN_LIB_ZEXT:
      {
	ZENRing Rg;

	Rg=ZENRingDef(R);
	if(ZENRingRecursiveAddPrc(Rg,prc))
	  return(ZERR);
	break;
      }
    }
  return(ZENRingAddPrc(R,prc));
}

#if 0
/* \TEX{
   \title{Suppressing precomputations.}
   \input{A finite ring |Rg| and precomputations |Prc|.} 
   \side{Structures already allocated by  |ZENRingRecursiveAddPrc()| are
   disallocated.}
   }
   */
void ZENRingRecursiveRmPrc(R, prc)
     ZENRing R;
     ZENPrc prc;{}
#endif
extern void _ZENRingRecursiveRmPrc(R, prc)
     ZENRing R;
     ZENPrc prc;
{
  switch(ZENRingArith(R))
    {
    case ZEN_LIB_ZEC:
      {
	int i;
	ZENRing *PR;

	PR=ZecRingSubrings(R);
	i=ZecRingNbS(R);	
	for(i--;i>=1;i--)
	  ZENRingRmPrc(PR[i],prc);
	ZENRingRecursiveRmPrc(PR[0],prc);
	break;
      }
    case ZEN_LIB_ZEXT:
      {
	ZENRing Rg;

	Rg=ZENRingDef(R);
	ZENRingRecursiveRmPrc(Rg,prc);
	break;
      }
    }
  ZENRingRmPrc(R,prc);
  return;
}

#if 0
/* \TEX{
   \title{Writing to file in a human comprehensive representation.}
   \input{A stream |file|, the level of ring, an allocated |ZENRing| |Rg|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{Printing a representation of |Rg| to |file|.}
   }
*/
int ZENRingPrintDebug(file, Rg)
     FILE *file;
     ZENRing Rg; {}
#endif
int _ZENRingPrintDebug(file, n, Rg)
     FILE *file;
     int n;
     ZENRing Rg;  
{
  extern char *_zen_library[];
  ZENPrc prc;
  
  if(fprintf(file,"[Level %d] Arithmetic: %s\n",n,_zen_library[ZENRingArith(Rg)])<0)
    {
      ZENSetError(ZEN_LIB_ZER, ZEN_PRINTTOFILE, ZEN_FPRINTF);
      return(ZERR);
    }

  prc=ZENRingPrcp(Rg);
  if(fprintf(file,"[Level %d] Precomputations: %d = (",n,(int)prc)<0)
    {
      ZENSetError(ZEN_LIB_ZER, ZEN_PRINTTOFILE, ZEN_FPRINTF);
      return(ZERR);
    }

#define PrcPrint(flag, str)							\
  if(prc & BlocExp2(flag))						\
    {									\
      if(fprintf(file,"%s|", str)<0)					\
	 {								\
	   ZENSetError(ZEN_LIB_ZER, ZEN_PRINTTOFILE, ZEN_FPRINTF);	\
	   return(ZERR);						\
	 }								\
      prc ^= BlocExp2(flag);						\
    }
  PrcPrint(ZENPRC_FINITE_FIELD,"ZENPRC_FINITE_FIELD");
  PrcPrint(ZENPRC_ELT_MULTIPLY, "ZENPRC_ELT_MULTIPLY");
  PrcPrint(ZENPRC_ELT_EXP, "ZENPRC_ELT_EXP");
  PrcPrint(ZENPRC_TRACE, "ZENPRC_TRACE");
  PrcPrint(ZENPRC_POLY_ROOTS_CANONICAL, "ZENPRC_POLY_ROOTS_CANONICAL");

  if(fprintf(file,"%d)\n",(int)prc)<0)
    {
      ZENSetError(ZEN_LIB_ZER, ZEN_PRINTTOFILE, ZEN_FPRINTF);
      return(ZERR);
    }
	   
  switch(ZENRingArith(Rg))
    {
    case ZEN_LIB_ZETAB:
    case ZEN_LIB_ZELOG:
    case ZEN_LIB_ZEM:
      break;
    case ZEN_LIB_ZEC:
      {
	int i;

	for(i=0;i<ZecRingNbS(Rg);i++)
	  {
	    fprintf(file,"[Level %d] Subring %d\n",n,i+1);
	    if(_ZENRingPrintDebug(file,n+1,ZecRingSubring(Rg,i)))
	      return(ZERR);
	  }
	break;
      }
    break;
    case ZEN_LIB_ZEXT:
      fprintf(file,"[Level %d] Polynomial: ",n);
      if(ZENPolyPrintToFile(file,ZENRingPol(Rg),10,ZENRingDef(Rg)))
	return(ZERR);
      fprintf(file,"\n");
      if(_ZENRingPrintDebug(file,n+1,ZENRingDef(Rg)))
	return(ZERR);
      break;

    default:
      fprintf(file,"[Level %d] Modulo: ",n);
      if(ZBNPrintToFile(file,ZENRingP(Rg),ZENRingPl(Rg),10))
	return(ZERR);
      fprintf(file,"\n");
      break;
    }
  if(ZENRingOrigin(Rg))
    {
      fprintf(file,"[Level %d] Cloned ring\n",n);
      if(_ZENRingPrintDebug(file,n+1,ZENRingOrigin(Rg)))
	return(ZERR);
    }
  return(0);
}
/* \TEX{
   \title{Printing to stderr.}
   \input{An allocated |ZENRing|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{Printing a representation of |R| to |stderr|.}    
   }
   */
int ZRP(R)
     ZENRing R;
{
  return(ZENRingPrintDebug(stderr,R));
}

