#include <stdio.h>
#include <zen.h>
#include <zenfact.h>

/* \TEX{
   \title{Polynomial factorization test}
   \input{A finite ring |Rg|, the number of tests and the degree of
   polynomials.}
   \output{0 if all the tests were ok, |ZENERR| otherwise.}
   } */
int ZENTestPolyIsNotIrreducible(Rg, nb, deg)
     ZENRing Rg;
     int nb;
     int deg;
{
  ZENPoly a,b,c;
  int i, dega,degb;
  double tt;

  printf("\nTesting ZENPolySetRandomIrreducible() and ZENPolyIsNotIrreducible()\n");
  tt = runtime();

  for(i = 0; i < nb; i++) {
    dega = 2 + (zbnrandom()%deg);
    if (!(a = ZENPolyCreate(dega, Rg))) {
      ZENError();
      printf("\n"); printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    
    if(ZENPolySetRandomIrreducible(a, dega, Rg)) {
      ZENError();
      printf("ZENPolySetRandomIrreducible() failed \n");
      return(ZERR);	
    }

    if(ZENPolyIsNotIrreducible(a,Rg)) {
      printf("\n"); printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    printf(".");fflush(stdout);

    if (!(b = ZENPolyCreate(dega*3, Rg))) {
      ZENError();
      printf("\n"); printf("The %dth test failed\n", i+1);
      return(ZERR);
    }

    ZENPolySetToXi(b,1,Rg);

    if (!(c = ZENPolyCreate(dega*2, Rg))) {
      ZENError();
      printf("\n"); printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    
    if (ZENPolyMultiply(c,a,b,Rg)) {
      ZENError();
      printf("\n"); printf("The %dth test failed\n", i+1);
      return(ZERR);
    }

    if(ZENPolyIsNotIrreducible(c,Rg)!= ZENPOLY_HAS_ZERO_ROOT) {
      printf("\n"); printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    printf(".");fflush(stdout);

    if (ZENPolySquare(c,a,Rg)) {
      ZENError();
      printf("\n"); printf("The %dth test failed\n", i+1);
      return(ZERR);
    }

    switch(ZENPolyIsNotIrreducible(c,Rg))
      {
      case ZENPOLY_IS_P_POWER:
	if((ZENRingPl(Rg)!=1)
	   ||(*(ZENRingP(Rg))!=2)) {
	  printf("\n"); printf("The %dth test failed\n", i+1);
	  return(ZERR);
	}
	break;

      case ZENPOLY_HAS_REPEATED_FACTORS:
	if((ZENRingPl(Rg)==1)
	   &&(*(ZENRingP(Rg))==2)) {
	  printf("\n"); printf("The %dth test failed\n", i+1);
	  return(ZERR);
	}
	break;

      default:
	printf("\n"); printf("The %dth test failed\n", i+1);
	return(ZERR);
      }
    
    printf(".");fflush(stdout);

    if (ZENPolyMultiply(b,c,a,Rg)) {
      ZENError();
      printf("\n"); printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    
    switch(ZENPolyIsNotIrreducible(b,Rg))
      {
      case ZENPOLY_IS_P_POWER:
	if((ZENRingPl(Rg)!=1)
	   ||(*(ZENRingP(Rg))!=3)) {
	  printf("\n"); printf("The %dth test failed\n", i+1);
	  return(ZERR);
	}
	break;
	
      case ZENPOLY_HAS_REPEATED_FACTORS:
	if((ZENRingPl(Rg)==1)
	   &&(*(ZENRingP(Rg))==3)) {
	  printf("\n"); printf("The %dth test failed\n", i+1);
	  return(ZERR);
	}
	break;
	
      default:
	printf("\n"); printf("The %dth test failed\n", i+1);
	return(ZERR);
      }
    printf(".");fflush(stdout);

    do {
      degb = 2 + (zbnrandom()%(dega*2-2));
      
      if(ZENPolySetRandomIrreducible(c, degb, Rg)) {
	ZENError();
	printf("ZENPolySetRandomIrreducible() failed \n");
	return(ZERR);	
      }
      if(ZENPolyIsNotIrreducible(a,Rg)) {
	printf("\n"); printf("The %dth test failed\n", i+1);
	return(ZERR);
      }
      printf("_");fflush(stdout);
    } while(ZENPolyAreEqual(c,a,Rg));

    printf(".");fflush(stdout);

    if (ZENPolyMultiply(b,c,a,Rg)) {
      ZENError();
      printf("\n"); printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    switch (ZENPolyIsNotIrreducible(b,Rg)) {
    case  ZENPOLY_IS_COMPOSED:
    case  ZENPOLY_HAS_REPEATED_FACTORS:
      break;
    default:
      printf("\n"); printf("The %dth test failed\n", i+1);
      return(ZERR);
    }

    ZENPolyFree(a,Rg);
    ZENPolyFree(b,Rg);
    ZENPolyFree(c,Rg);

    printf("*");fflush(stdout);
   
  }
  printf("\nZENPolyIsNotIrreducible() successfully tested in %.2f s\n",
	 (runtime()-tt)/1000.);
  return(0);
}

/* \TEX{
   \title{Polynomial roots test}
   \input{A finite ring |Rg|, the number of tests and the degree of
   polynomials.}
   \output{0 if all the tests were ok, |ZENERR| otherwise.}
   } */
int ZENTestPolyRoots(Rg, nb, deg)
     ZENRing Rg;
     int nb;
     int deg;
{
  ZENPoly a,b;
  ZENElt elt;
  int i, j, dega;
  double tt;

  printf("\nTesting ZENPolyRoots()\n");
  tt = runtime();
       
  for(i = 0; i < nb; i++) {
    dega = 2 + (zbnrandom()%deg);
    
    if (!(a = ZENPolyCreate(dega, Rg))) {
      ZENError();
      printf("\n"); printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    if (!(b = ZENPolyCreate(dega, Rg))) {
      ZENError();
      printf("\n"); printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    ZENPolySetRandom(a, dega, Rg);

    if(ZENPolyRoots(b, a,Rg)) {
      printf("\n"); printf("The %dth test failed\n", i+1);
      return(ZERR);
     }

    for(j = 0; j < ZENPolyDeg(b, Rg); j++) {
      elt = ZENPolyGetCoeffPtr(b, j, Rg);
      ZENPolyEval(elt, a, elt, Rg);
      if (!ZENEltIsZero(elt, Rg)) {
	printf("\n"); printf("The %dth test failed\n", i+1);
	return(ZERR);
      }
    }
    
    printf(".");fflush(stdout);
    ZENPolyFree(a,Rg);
    ZENPolyFree(b,Rg);
  }
  printf("\nZENPolyRoots() successfully tested in %.2f s\n",
	 (runtime()-tt)/1000.);
  return(0);
}

/* \TEX{
   \title{Polynomial factorization test}
   \input{A finite ring |Rg|, the number of tests and the degree of
   polynomials.}
   \output{0 if all the tests were ok, |ZENERR| otherwise.}
   } */
int ZENTestPolyFactor(Rg, nb, deg)
     ZENRing Rg;
     int nb;
     int deg;
{
  ZENPoly a, b;
  int i, j, dega;
  double tt;
  ZENPolyFact lfact;


  printf("\nTesting ZENPolyFactor()\n");
  tt = runtime();

  for(i = 0; i < nb; i++) {
    ZENPolyFactCreate(lfact, Rg);

    dega = 2 + (zbnrandom()%deg);
    
    if (!(a = ZENPolyCreate(dega, Rg))) {
      ZENError();
      printf("\n"); printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    ZENPolySetRandom(a, dega, Rg);

    if (!(b = ZENPolyCreate(dega, Rg))) {
      ZENError();
      printf("\n"); printf("The %dth test failed\n", i+1);
      return(ZERR);
    }

    if(ZENPolyFactor(lfact, a, Rg)) {
      printf("\n"); printf("The %dth test failed\n", i+1);
      return(ZERR);
     }

    deg = 0;
    ZENPolyAssign(b, a, Rg);
    for(j = 0; j < lfact->nb; j++) {
      deg += lfact->LP[j].m*ZENPolyDeg(lfact->LP[j].P, Rg);


      if (ZENPolyDivide(b, b, a, lfact->LP[j].P, Rg) != ZEN_HAS_INVERSE) {
	printf("\n"); printf("The %dth test failed\n", i+1);
	return(ZERR);
      }
      if (!ZENPolyIsZero(b, Rg)) {
	printf("\n"); printf("The %dth test failed\n", i+1);
	return(ZERR);
      }
    }
    if (deg != dega) {
      printf("\n"); printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    
    printf(".");fflush(stdout);
    ZENPolyFactFree(lfact, Rg);
    ZENPolyFree(a,Rg);
    ZENPolyFree(b,Rg);
  }
  printf("\nZENPolyFactor() successfully tested in %.2f s\n",
	 (runtime()-tt)/1000.);
  return(0);
}



