#ifndef __ZENFACT__
#  define __ZENFACT__
#define ZENFACT_VERSION 2.1

/* Some definitions */
#define ZENPOLY_IS_IRREDUCIBLE       ZEN_HAS_INVERSE
#define ZENPOLY_HAS_ZERO_ROOT          2
#define ZENPOLY_HAS_ONE_ROOT           3
#define ZENPOLY_HAS_REPEATED_FACTORS   4
#define ZENPOLY_IS_P_POWER             5
#define ZENPOLY_IS_COMPOSED            6
#define ZENPOLY_IS_SURELY_IRREDUCIBLE  128 /* This is needed by
					      ConstructBerlekamp and means
					      ZENPOLY_IS_IRREDUCIBLE
					      Do NOT use. */

/* \TEX{
   \Chapter{Enumerating the functions of \zenfact}
   \Section{Roots of polynomials.}

   \input{roots}
   }
   */
#define ZENPolyRoots(roots,P,Rg) _ZENPolyRoots(roots,P,Rg)
extern int _ZENPolyRoots ___((ZENPoly,ZENPoly,ZENRing));

#define ZENPolyRootsEDF(roots,P,Rg) _ZENPolyRootsEDF(roots,P,Rg)
extern int _ZENPolyRootsEDF ___((ZENPoly,ZENPoly,ZENRing));

#define ZENPolyRootsBerlekampEvenChar(roots,P,Rg) \
_ZENPolyRootsBerlekampEvenChar(roots,P,Rg)
extern int _ZENPolyRootsBerlekampEvenChar ___((ZENPoly,ZENPoly,ZENRing));

#define ZENPolyRootsBerlekampOddChar(roots,P,Rg) \
_ZENPolyRootsBerlekampOddChar(roots,P,Rg)
extern int _ZENPolyRootsBerlekampOddChar ___((ZENPoly,ZENPoly,ZENRing));

#define ZENPolyOneRootEDF(root,P,Rg) _ZENPolyOneRootEDF(&(root),P,Rg)
extern int _ZENPolyOneRootEDF ___((ZENElt *,ZENPoly,ZENRing));

#define ZENPolyOneRootBerlekampEvenChar(root,P,Rg) \
_ZENPolyOneRootBerlekampEvenChar(&(root),P,Rg)
extern int _ZENPolyOneRootBerlekampEvenChar ___((ZENElt *,ZENPoly,ZENRing));

#define ZENPolyOneRootBerlekampOddChar(root,P,Rg) \
_ZENPolyOneRootBerlekampOddChar(&(root),P,Rg)
extern int _ZENPolyOneRootBerlekampOddChar ___((ZENElt *,ZENPoly,ZENRing));

#define ZENEltPRoot(a,b,Rg) _ZENEltPRoot(&(a),b,Rg)
extern int _ZENEltPRoot ___((ZENElt*, ZENElt, ZENRing));

#define ZENEltDRoot(R, d, A, Rg) _ZENEltDRoot(&(R), d, A, Rg)
extern int _ZENEltDRoot ___((ZENElt *, BigNumDigit, ZENElt, ZENRing));

#define ZENEltIsADRoot(d, A, Rg) _ZENEltIsADRoot(d, A, Rg)
extern int _ZENEltIsADRoot ___((BigNumDigit, ZENElt, ZENRing));

/* \TEX{
   \Section{Polynomial factorization.}
   
   \input{factors}
   }
   */


typedef struct zenpolylist {
  ZENPoly P, Xq;
  int m, k;
} ZENPolyList;

typedef struct zenpolyfact{
  int nb;
  int lgt;
  ZENPolyList *LP;
} *ZENPolyFact;


#define ZENPolyListFree(list, Rg) if (1) {ZENPolyFT(list.P,Rg); ZENPolyFT(list.Xq,Rg);} else

#define ZENPolyFactCreate(lfact,Rg) ((lfact)=NULL)

#define ZENPolyFactFree(lfact,Rg) \
_ZENPolyFactFree(&(lfact),Rg)
extern void _ZENPolyFactFree ___((ZENPolyFact*, ZENRing));

#define ZENPolyFactAppend(lfact,list,Rg) \
_ZENPolyFactAppend(&(lfact),list,Rg)
extern int _ZENPolyFactAppend ___((ZENPolyFact*, ZENPolyList, ZENRing));

#define ZENPolyFactRemove(list, lfact, index, Rg) \
_ZENPolyFactRemove(&(list), &(lfact), index, Rg)
extern int _ZENPolyFactRemove ___((ZENPolyList*, ZENPolyFact *, int, ZENRing));

#define ZENPolyFactor(lfact, P, Rg) \
_ZENPolyFactor(&(lfact), P, Rg)
extern int _ZENPolyFactor ___((ZENPolyFact*, ZENPoly, ZENRing));

#define ZENPolySFFactor(lfact, P, Rg) \
_ZENPolySFFactor(&(lfact), P, Rg)
extern int _ZENPolySFFactor ___((ZENPolyFact*, ZENPoly, ZENRing));

#define ZENPolyDDFactor(lfact, P, Rg) \
_ZENPolyDDFactor(&(lfact), P, Rg)
extern int _ZENPolyDDFactor ___((ZENPolyFact*, ZENPoly, ZENRing));

#define ZENPolyEDFactor(lfact, P, Xq, k, Rg) \
_ZENPolyEDFactor(&(lfact), P, Xq, k, Rg)
extern int _ZENPolyEDFactor ___((ZENPolyFact*, ZENPoly, ZENPoly, int, ZENRing));

#define ZENPolyThiongLy(P, XqmodP, k, Rg) _ZENPolyThiongLy(P, XqmodP, k, Rg)
extern ZENPoly  _ZENPolyThiongLy ___((ZENPoly, ZENPoly, int, ZENRing));

#define ZENPolyCamion(P, k, Rg) _ZENPolyCamion(P, k, Rg)
extern ZENPoly  _ZENPolyCamion ___((ZENPoly, int, ZENRing));

#define ZENShoopTrace(T, Q, k, XqmodP, Phi) _ZENShoopTrace(T, Q, k, XqmodP, Phi)
extern int  _ZENShoopTrace ___((ZENElt, ZENElt, int, ZENElt, ZENRing));

#define ZENPolyPRoot(P, Q, Rg) _ZENPolyPRoot(P, Q, Rg)
extern int _ZENPolyPRoot ___((ZENPoly,ZENPoly,ZENRing));

/*
  \TEX{
  \Section{Irreducible polynomials.}
  } */

#if 0
/*
   \TEX{
     \title{Testing irreducibility of a polynomial}
     \input{A polynomial |PX| defined over a finite ring |Rg|.}
     \output{\begin{description}
       \item[|ZENERR|] if an error occured,
       \item[|ZEN_NO_INVERSE|] if an inverse was impossible to compute,
       \item[|ZENPOLY_HAS_ZERO_ROOT|] if the polynomial has zero as root,
       \item[|ZENPOLY_HAS_ONE_ROOT|] if the polynomial has one as root, 
       \item[|ZENPOLY_HAS_REPEATED_FACTORS|] if the polynomial has repeated factors,
       \item[|ZENPOLY_IS_P_POWER|] if $|P|(X)= g(X)^p$, with $p$ the
         characteristic of the |ZENRing|,
       \item[|ZENPOLY_IS_IRREDUCIBLE|] if the polynomial is irreducible,
       \item[|ZENPOLY_IS_COMPOSED|] otherwise.
       \end{description}}
     \note{Valid only in finite fields. The Berlekamp's algorithm is used}
   } 
   */
int ZENPolyIsNotIrreducible(PX,Rg)
     ZENPoly PX;
     ZENRing Rg; {}
#endif
#define ZENPolyIsNotIrreducible(PX,Rg) _ZENPolyIsNotIrreducible(PX,Berlekamp,Rg)
extern int _ZENPolyIsNotIrreducible ___((ZENPoly,int (*) ___((ZENPoly,ZENRing)), ZENRing));
extern int Berlekamp ___((ZENPoly,ZENRing));
extern int DDF ___((ZENPoly,ZENRing));

/*
  \TEX{
  \input{berlekamp}
   We wrote two functions to find irreducible polynomials.

  }
  */
#if 0
/* \TEX{
   \title{Setting a polynomial to an irreducible random one}
   \input{A polynomial, a degree, a |ZENRing|.}
   \output{|ZENERR| if an error occured, 0 otherwise}
   \side{The polynomial is set randomly to a monic polynomial of
   given degree that is irreducible over |R|.}
   \note{This procedure may take a while. It is valid only in finite fields. The Berlekamp's algorithm is used.}
   }
   */
int ZENPolySetRandomIrreducible(P,deg,R)
     ZENPoly P;
     int deg;
     ZENRing R;{}
#endif
#define ZENPolySetRandomIrreducible(p,i,R) _ZENPolySetRandomIrreducible(p,i,Berlekamp,R) 
extern int _ZENPolySetRandomIrreducible ___((ZENPoly, int, int (*) ___((ZENPoly,ZENRing)), ZENRing));

#if 0
/*
  \TEX{
  \title{Finding the smallest polynomial $f(X)$ such that $X^d+f(X)$ is an irreducible polynomial.}
  \input{A degree |d| and an allocated polynomial |P| of size at
  least |d| of a |ZENRing| |R|.}
  \output{-1 if an error occured, 0 otherwise}
  \side{The polynomial is set to a polynomial of given degree that is irreducible over |R|}
  \note{This procedure may take a while. It uses Berlekamp's algorithm.}
  } 
*/
int ZENPolySetSmallestIrreducible(P,deg,R)
     ZENPoly P;
     int deg;
     ZENRing R;{}
#endif
#define ZENPolySetSmallestIrreducible(P,d,R) _ZENPolySetSmallestIrreducible(P,d,Berlekamp,R)
extern int _ZENPolySetSmallestIrreducible ___((ZENPoly, int, int (*) ___((ZENPoly,ZENRing)), ZENRing));

#if 0
/*
  \TEX{
  \title{Finding a small random polynomial $f(X)$ such that $X^d+f(X)$ is an irreducible polynomial.}
  \input{A degree |d| and an allocated polynomial |P| of size at
  least |d| of a |ZENRing| |R|.}
  \output{-1 if an error occured, 0 otherwise}
  \side{The polynomial is set to a polynomial of given degree that is irreducible over |R|}
  \note{This procedure may take a while. It uses Berlekamp's algorithm.}
  } 
*/
int ZENPolySetGoodIrreducible(P,deg,R)
     ZENPoly P;
     int deg;
     ZENRing R;{}
#endif
#define ZENPolySetGoodIrreducible(P,d,R) _ZENPolySetGoodIrreducible(P,d,Berlekamp,R)
extern int _ZENPolySetGoodIrreducible ___((ZENPoly, int, int (*) ___((ZENPoly,ZENRing)), ZENRing));

/*
  \TEX{
  \input{irreduc}

  \Section{Pseudo-prime integers.}

  For integers, only a pseudo primality test based on a Miller Rabin
  algorithm was implemented.

  \input{psprime}
  } */
#define ZENMillerRabin(Rg, e) _ZENMillerRabin(Rg, e)
extern int _ZENMillerRabin ___((ZENRing,BigNumDigit));

#define ZENRingIsAPseudoFF(R) _ZENRingIsAPseudoFF(R)
extern int _ZENRingIsAPseudoFF ___((ZENRing));

extern int ZBNIsPrime ___((BigNum,BigNumLength));
extern int ZBNNextPrime ___((BigNum *,BigNumLength *));

/*
   \TEX{
   \Section{Finite field construction}
   We provide high-level functions to easily build a representation of any finite field. This representation is not always the most efficient, but compromise should be good enough for most purposes.
   } 
   */
#define ZENFieldAlloc(R,q,ql,d) (((R)=_ZENFieldCreate(q,ql,d))==NULL ? ZERR : 0)
#define ZENFieldCreate(q,ql,d) _ZENFieldCreate(q,ql,d)
extern ZENRing _ZENFieldCreate ___((BigNum, BigNumLength, int));

#define ZENRingDefine(in,out) _ZENRingDefine(in,out)
extern ZENRing _ZENRingDefine ___((FILE *, FILE *));

#define ZENRingRecursiveAddPrc(R, prc) _ZENRingRecursiveAddPrc(R, prc)
extern int _ZENRingRecursiveAddPrc ___((ZENRing, ZENPrc));

#define ZENRingRecursiveRmPrc(R, prc) _ZENRingRecursiveRmPrc(R, prc)
extern void _ZENRingRecursiveRmPrc ___((ZENRing, ZENPrc));

#define ZENRingPrintDebug(file, Rg) _ZENRingPrintDebug(file, 0, Rg) 
extern int _ZENRingPrintDebug ___((FILE *, int, ZENRing));

/*
   \TEX{
   \input{ring}   
   } 
   */



#endif
